@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2010-2012 Jeroen Frijters
+  Copyright (C) 2010-2013 Jeroen Frijters
   Copyright (C) 2011 Marek Safar
 
   This software is provided 'as-is', without any express or implied
@@ -32,7 +32,7 @@ namespace IKVM.Reflection
 	struct ParsedAssemblyName
 	{
 		internal string Name;
-		internal Version Version;
+		internal string Version;
 		internal string Culture;
 		internal string PublicKeyToken;
 		internal bool? Retargetable;
@@ -50,6 +50,16 @@ namespace IKVM.Reflection
 
 	static class Fusion
 	{
+		static readonly Version FrameworkVersion = new Version(4, 0, 0, 0);
+		static readonly Version FrameworkVersionNext = new Version(4, 1, 0, 0);
+		static readonly Version SilverlightVersion = new Version(2, 0, 5, 0);
+		static readonly Version SilverlightVersionMinimum = new Version(2, 0, 0, 0);
+		static readonly Version SilverlightVersionMaximum = new Version(5, 9, 0, 0);
+		const string PublicKeyTokenEcma = "b77a5c561934e089";
+		const string PublicKeyTokenMicrosoft = "b03f5f7f11d50a3a";
+		const string PublicKeyTokenSilverlight = "7cec85d7bea7798e";
+		const string PublicKeyTokenWinFX = "31bf3856ad364e35";
+
 		internal static bool CompareAssemblyIdentityNative(string assemblyIdentity1, bool unified1, string assemblyIdentity2, bool unified2, out AssemblyComparisonResult result)
 		{
 			bool equivalent;
@@ -80,19 +90,32 @@ namespace IKVM.Reflection
 				}
 			}
 
-			bool partial = IsPartial(name1);
-
-			if ((partial && unified1) || IsPartial(name2))
+			Version version1;
+			Version version2;
+			if (!ParseVersion(name1.Version, out version1) || !ParseVersion(name2.Version, out version2))
 			{
 				result = AssemblyComparisonResult.NonEquivalent;
 				throw new ArgumentException();
 			}
-			if (!name1.Name.Equals(name2.Name, StringComparison.InvariantCultureIgnoreCase))
+
+			bool partial = IsPartial(name1, version1);
+
+			if (partial && name1.Retargetable.HasValue)
+			{
+				result = AssemblyComparisonResult.NonEquivalent;
+				throw new System.IO.FileLoadException();
+			}
+			if ((partial && unified1) || IsPartial(name2, version2))
+			{
+				result = AssemblyComparisonResult.NonEquivalent;
+				throw new ArgumentException();
+			}
+			if (!name1.Name.Equals(name2.Name, StringComparison.OrdinalIgnoreCase))
 			{
 				result = AssemblyComparisonResult.NonEquivalent;
 				return false;
 			}
-			if (name1.Name.Equals("mscorlib", StringComparison.InvariantCultureIgnoreCase))
+			if (name1.Name.Equals("mscorlib", StringComparison.OrdinalIgnoreCase))
 			{
 				result = AssemblyComparisonResult.EquivalentFullMatch;
 				return true;
@@ -100,41 +123,89 @@ namespace IKVM.Reflection
 			if (partial && name1.Culture == null)
 			{
 			}
-			else if (!name1.Culture.Equals(name2.Culture, StringComparison.InvariantCultureIgnoreCase))
+			else if (!name1.Culture.Equals(name2.Culture, StringComparison.OrdinalIgnoreCase))
 			{
 				result = AssemblyComparisonResult.NonEquivalent;
 				return false;
 			}
+
+			if (!name1.Retargetable.GetValueOrDefault() && name2.Retargetable.GetValueOrDefault())
+			{
+				result = AssemblyComparisonResult.NonEquivalent;
+				return false;
+			}
+
+			// HACK handle the case "System.Net, Version=4.0.0.0, Culture=neutral, PublicKeyToken=7cec85d7bea7798e, Retargetable=Yes"
+			// compared with "System.Net, Version=4.0.0.0, Culture=neutral, PublicKeyToken=7cec85d7bea7798e, Retargetable=No"
+			if (name1.PublicKeyToken == name2.PublicKeyToken
+				&& version1 != null
+				&& name1.Retargetable.GetValueOrDefault()
+				&& !name2.Retargetable.GetValueOrDefault()
+				&& GetRemappedPublicKeyToken(ref name1, version1) != null)
+			{
+				name1.Retargetable = false;
+			}
+
+			string remappedPublicKeyToken1 = null;
+			string remappedPublicKeyToken2 = null;
+			if (version1 != null && (remappedPublicKeyToken1 = GetRemappedPublicKeyToken(ref name1, version1)) != null)
+			{
+				name1.PublicKeyToken = remappedPublicKeyToken1;
+				version1 = FrameworkVersion;
+			}
+			if ((remappedPublicKeyToken2 = GetRemappedPublicKeyToken(ref name2, version2)) != null)
+			{
+				name2.PublicKeyToken = remappedPublicKeyToken2;
+				version2 = FrameworkVersion;
+			}
+			if (name1.Retargetable.GetValueOrDefault())
+			{
+				if (name2.Retargetable.GetValueOrDefault())
+				{
+					if (remappedPublicKeyToken1 != null ^ remappedPublicKeyToken2 != null)
+					{
+						result = AssemblyComparisonResult.NonEquivalent;
+						return false;
+					}
+				}
+				else if (remappedPublicKeyToken1 == null || remappedPublicKeyToken2 != null)
+				{
+					result = AssemblyComparisonResult.Unknown;
+					return false;
+				}
+			}
+
+			bool fxUnified = false;
+			bool versionMatch = version1 == version2;
+			if (IsFrameworkAssembly(name1))
+			{
+				fxUnified |= !versionMatch;
+				version1 = FrameworkVersion;
+			}
+			if (IsFrameworkAssembly(name2) && version2 < FrameworkVersionNext)
+			{
+				fxUnified |= !versionMatch;
+				version2 = FrameworkVersion;
+			}
+
 			if (IsStrongNamed(name2))
 			{
-				if (partial && name1.PublicKeyToken == null)
+				if (name1.PublicKeyToken != null && name1.PublicKeyToken != name2.PublicKeyToken)
 				{
-				}
-				else if (name1.PublicKeyToken != name2.PublicKeyToken)
-				{
-					if (HasPclRuntimeRemapping (name1, name2)) {
-						result = AssemblyComparisonResult.EquivalentFXUnified;
-						return true;
-					}
 					result = AssemblyComparisonResult.NonEquivalent;
 					return false;
 				}
-				if (partial && name1.Version == null)
+				else if (version1 == null)
 				{
 					result = AssemblyComparisonResult.EquivalentPartialMatch;
 					return true;
 				}
-				else if (IsFrameworkAssembly(name2) || IsFacadeAssembly(name2))
-				{
-					result = partial ? AssemblyComparisonResult.EquivalentPartialFXUnified : AssemblyComparisonResult.EquivalentFXUnified;
-					return true;
-				}
-				else if (name1.Version.Revision == -1 || name2.Version.Revision == -1)
+				else if (version1.Revision == -1 || version2.Revision == -1)
 				{
 					result = AssemblyComparisonResult.NonEquivalent;
 					throw new ArgumentException();
 				}
-				else if (name1.Version < name2.Version)
+				else if (version1 < version2)
 				{
 					if (unified2)
 					{
@@ -147,7 +218,7 @@ namespace IKVM.Reflection
 						return false;
 					}
 				}
-				else if (name1.Version > name2.Version)
+				else if (version1 > version2)
 				{
 					if (unified1)
 					{
@@ -160,6 +231,11 @@ namespace IKVM.Reflection
 						return false;
 					}
 				}
+				else if (!versionMatch || fxUnified)
+				{
+					result = partial ? AssemblyComparisonResult.EquivalentPartialFXUnified : AssemblyComparisonResult.EquivalentFXUnified;
+					return true;
+				}
 				else
 				{
 					result = partial ? AssemblyComparisonResult.EquivalentPartialMatch : AssemblyComparisonResult.EquivalentFullMatch;
@@ -178,56 +254,10 @@ namespace IKVM.Reflection
 			}
 		}
 
-		#region PCL Runtime Remapping
-
-		// keep this in sync with the key_remap_table in mono/metadata/assembly.c
-
-		const string SILVERLIGHT_KEY = "7cec85d7bea7798e";
-		const string WINFX_KEY = "31bf3856ad364e35";
-		const string ECMA_KEY = "b77a5c561934e089";
-		const string MSFINAL_KEY = "b03f5f7f11d50a3a";
-
-		static readonly string[,] key_remap_table = new string[,] {
-			{ "Microsoft.CSharp", WINFX_KEY, MSFINAL_KEY },
-			{ "System", SILVERLIGHT_KEY, ECMA_KEY },
-			{ "System.ComponentModel.Composition", WINFX_KEY, ECMA_KEY },
-			{ "System.ComponentModel.DataAnnotations", "ddd0da4d3e678217", WINFX_KEY },
-			{ "System.Core", SILVERLIGHT_KEY, ECMA_KEY },
-			// FIXME: MS uses MSFINAL_KEY for .NET 4.5
-			{ "System.Net", SILVERLIGHT_KEY, MSFINAL_KEY },
-			{ "System.Numerics", WINFX_KEY, MSFINAL_KEY },
-			{ "System.Runtime.Serialization", SILVERLIGHT_KEY, ECMA_KEY },
-			{ "System.ServiceModel", WINFX_KEY, ECMA_KEY },
-			{ "System.ServiceModel.Web", SILVERLIGHT_KEY, WINFX_KEY },
-			{ "System.Windows", SILVERLIGHT_KEY, MSFINAL_KEY },
-			{ "System.Xml", SILVERLIGHT_KEY, ECMA_KEY },
-			{ "System.Xml.Linq", WINFX_KEY, ECMA_KEY },
-			{ "System.Xml.Serialization", WINFX_KEY, ECMA_KEY }
-		};
-
-		static bool HasPclRuntimeRemapping(ParsedAssemblyName name1, ParsedAssemblyName name2)
-		{
-			if (name1.Name != name2.Name)
-				return false;
-
-			for (int i = 0; i < key_remap_table.GetLength (0); i++) {
-				if (!name1.Name.Equals (key_remap_table [i,0]))
-					continue;
-				if (name1.PublicKeyToken.Equals (key_remap_table [i, 1]) &&
-				    name2.PublicKeyToken.Equals (key_remap_table [i, 2]))
-					return true;
-			}
-
-			return false;
-		}
-
-		#endregion
-
 		static bool IsFrameworkAssembly(ParsedAssemblyName name)
 		{
-			// A list of FX assemblies which require some form of remapping
-			// When 4.0 + 1 version  is release, assemblies introduced in v4.0
-			// will have to be added
+			// Framework assemblies use different unification rules, so when
+			// a new framework is released the new assemblies need to be added.
 			switch (name.Name)
 			{
 				case "System":
@@ -240,31 +270,90 @@ namespace IKVM.Reflection
 				case "System.Data.Services.Client":
 				case "System.IdentityModel":
 				case "System.IdentityModel.Selectors":
+				case "System.IO.Compression":
+				case "System.Numerics":
+				case "System.Reflection.Context":
 				case "System.Runtime.Remoting":
 				case "System.Runtime.Serialization":
+				case "System.Runtime.WindowsRuntime":
+				case "System.Runtime.WindowsRuntime.UI.Xaml":
 				case "System.ServiceModel":
 				case "System.Transactions":
 				case "System.Windows.Forms":
 				case "System.Xml":
 				case "System.Xml.Linq":
-					return name.PublicKeyToken == "b77a5c561934e089";
+				case "System.Xml.Serialization":
+					return name.PublicKeyToken == PublicKeyTokenEcma;
 
+				case "Microsoft.CSharp":
+				case "Microsoft.VisualBasic":
+				case "System.Collections":
+				case "System.Collections.Concurrent":
+				case "System.ComponentModel":
+				case "System.ComponentModel.Annotations":
+				case "System.ComponentModel.EventBasedAsync":
 				case "System.Configuration":
 				case "System.Configuration.Install":
 				case "System.Design":
+				case "System.Diagnostics.Contracts":
+				case "System.Diagnostics.Debug":
+				case "System.Diagnostics.Tools":
+				case "System.Diagnostics.Tracing":
 				case "System.DirectoryServices":
 				case "System.Drawing":
 				case "System.Drawing.Design":
+				case "System.Dynamic.Runtime":
 				case "System.EnterpriseServices":
+				case "System.Globalization":
+				case "System.IO":
+				case "System.Linq":
+				case "System.Linq.Expressions":
+				case "System.Linq.Parallel":
+				case "System.Linq.Queryable":
 				case "System.Management":
 				case "System.Messaging":
+				case "System.Net":
+				case "System.Net.Http":
+				case "System.Net.Http.Rtc":
+				case "System.Net.NetworkInformation":
+				case "System.Net.Primitives":
+				case "System.Net.Requests":
+				case "System.ObjectModel":
+				case "System.Reflection":
+				case "System.Reflection.Extensions":
+				case "System.Reflection.Primitives":
+				case "System.Resources.ResourceManager":
+				case "System.Runtime":
+				case "System.Runtime.Extensions":
+				case "System.Runtime.InteropServices":
+				case "System.Runtime.InteropServices.WindowsRuntime":
+				case "System.Runtime.Numerics":
 				case "System.Runtime.Serialization.Formatters.Soap":
+				case "System.Runtime.Serialization.Json":
+				case "System.Runtime.Serialization.Primitives":
+				case "System.Runtime.Serialization.Xml":
 				case "System.Security":
+				case "System.Security.Principal":
+				case "System.ServiceModel.Duplex":
+				case "System.ServiceModel.Http":
+				case "System.ServiceModel.NetTcp":
+				case "System.ServiceModel.Primitives":
+				case "System.ServiceModel.Security":
 				case "System.ServiceProcess":
+				case "System.Text.Encoding":
+				case "System.Text.Encoding.Extensions":
+				case "System.Text.RegularExpressions":
+				case "System.Threading":
+				case "System.Threading.Tasks":
+				case "System.Threading.Tasks.Parallel":
 				case "System.Web":
 				case "System.Web.Mobile":
 				case "System.Web.Services":
-					return name.PublicKeyToken == "b03f5f7f11d50a3a";
+				case "System.Windows":
+				case "System.Xml.ReaderWriter":
+				case "System.Xml.XDocument":
+				case "System.Xml.XmlSerializer":
+					return name.PublicKeyToken == PublicKeyTokenMicrosoft;
 
 				case "System.ComponentModel.DataAnnotations":
 				case "System.ServiceModel.Web":
@@ -273,119 +362,77 @@ namespace IKVM.Reflection
 				case "System.Web.Extensions.Design":
 				case "System.Web.DynamicData":
 				case "System.Web.Routing":
-					return name.PublicKeyToken == "31bf3856ad364e35";
+					return name.PublicKeyToken == PublicKeyTokenWinFX;
 			}
 
 			return false;
 		}
 
-		static bool IsFacadeAssembly (ParsedAssemblyName name)
+		static string GetRemappedPublicKeyToken(ref ParsedAssemblyName name, Version version)
 		{
-			switch (name.Name) {
-			case "System.Collections":
-			case "System.Collections.Concurrent":
-			case "System.ComponentModel":
-			case "System.ComponentModel.Annotations":
-			case "System.ComponentModel.EventBasedAsync":
-			case "System.Diagnostics.Contracts":
-			case "System.Diagnostics.Debug":
-			case "System.Diagnostics.Tools":
-			case "System.Dynamic.Runtime":
-			case "System.Globalization":
-			case "System.IO":
-			case "System.Linq":
-			case "System.Linq.Expressions":
-			case "System.Linq.Parallel":
-			case "System.Linq.Queryable":
-			case "System.Net.NetworkInformation":
-			case "System.Net.Primitives":
-			case "System.Net.Requests":
-			case "System.ObjectModel":
-			case "System.Reflection":
-			case "System.Reflection.Extensions":
-			case "System.Reflection.Primitives":
-			case "System.Resources.ResourceManager":
-			case "System.Runtime":
-			case "System.Runtime.Extensions":
-			case "System.Runtime.InteropServices":
-			case "System.Runtime.Numerics":
-			case "System.Runtime.Serialization.Json":
-			case "System.Runtime.Serialization.Primitives":
-			case "System.Runtime.Serialization.Xml":
-			case "System.Security.Principal":
-			case "System.ServiceModel.Http":
-			case "System.ServiceModel.Primitives":
-			case "System.Text.Encoding":
-			case "System.Text.Encoding.Extensions":
-			case "System.Text.RegularExpressions":
-			case "System.Threading":
-			case "System.Threading.Tasks":
-			case "System.Threading.Tasks.Parallel":
-			case "System.Xml.ReaderWriter":
-			case "System.Xml.XDocument":
-			case "System.Xml.XmlSerializer":
-				return name.PublicKeyToken == "b03f5f7f11d50a3a";
-			default:
-				return false;
+			if (name.Retargetable.GetValueOrDefault() && version < SilverlightVersion)
+			{
+				return null;
 			}
+			if (name.PublicKeyToken == "ddd0da4d3e678217" && name.Name == "System.ComponentModel.DataAnnotations" && name.Retargetable.GetValueOrDefault())
+			{
+				return PublicKeyTokenWinFX;
+			}
+			if (SilverlightVersionMinimum <= version && version <= SilverlightVersionMaximum)
+			{
+				switch (name.PublicKeyToken)
+				{
+					case PublicKeyTokenSilverlight:
+						switch (name.Name)
+						{
+							case "System":
+							case "System.Core":
+								return PublicKeyTokenEcma;
+						}
+						if (name.Retargetable.GetValueOrDefault())
+						{
+							switch (name.Name)
+							{
+								case "System.Runtime.Serialization":
+								case "System.Xml":
+									return PublicKeyTokenEcma;
+								case "System.Net":
+								case "System.Windows":
+									return PublicKeyTokenMicrosoft;
+								case "System.ServiceModel.Web":
+									return PublicKeyTokenWinFX;
+							}
+						}
+						break;
+					case PublicKeyTokenWinFX:
+						switch (name.Name)
+						{
+							case "System.ComponentModel.Composition":
+								return PublicKeyTokenEcma;
+						}
+						if (name.Retargetable.GetValueOrDefault())
+						{
+							switch (name.Name)
+							{
+								case "Microsoft.CSharp":
+									return PublicKeyTokenMicrosoft;
+								case "System.Numerics":
+								case "System.ServiceModel":
+								case "System.Xml.Serialization":
+								case "System.Xml.Linq":
+									return PublicKeyTokenEcma;
+							}
+						}
+						break;
+				}
+			}
+			return null;
 		}
 
 		internal static ParseAssemblyResult ParseAssemblySimpleName(string fullName, out int pos, out string simpleName)
 		{
-			StringBuilder sb = new StringBuilder();
 			pos = 0;
-			simpleName = null;
-			while (pos < fullName.Length && char.IsWhiteSpace(fullName[pos]))
-			{
-				pos++;
-			}
-			char quoteOrComma = ',';
-			if (pos < fullName.Length && (fullName[pos] == '\"' || fullName[pos] == '\''))
-			{
-				quoteOrComma = fullName[pos++];
-			}
-			while (pos < fullName.Length)
-			{
-				char ch = fullName[pos++];
-				if (ch == '\\')
-				{
-					if (pos == fullName.Length)
-					{
-						return ParseAssemblyResult.GenericError;
-					}
-					ch = fullName[pos++];
-					if (ch == '\\')
-					{
-						return ParseAssemblyResult.GenericError;
-					}
-				}
-				else if (ch == quoteOrComma)
-				{
-					if (ch != ',')
-					{
-						while (pos != fullName.Length)
-						{
-							ch = fullName[pos++];
-							if (ch == ',')
-							{
-								break;
-							}
-							if (!char.IsWhiteSpace(ch))
-							{
-								return ParseAssemblyResult.GenericError;
-							}
-						}
-					}
-					break;
-				}
-				else if (ch == '=' || (quoteOrComma == ',' && (ch == '\'' || ch == '"')))
-				{
-					return ParseAssemblyResult.GenericError;
-				}
-				sb.Append(ch);
-			}
-			simpleName = sb.ToString().Trim();
-			if (simpleName.Length == 0)
+			if (!TryParse(fullName, ref pos, out simpleName) || simpleName.Length == 0)
 			{
 				return ParseAssemblyResult.GenericError;
 			}
@@ -396,12 +443,88 @@ namespace IKVM.Reflection
 			return ParseAssemblyResult.OK;
 		}
 
+		private static bool TryParse(string fullName, ref int pos, out string value)
+		{
+			value = null;
+			StringBuilder sb = new StringBuilder();
+			while (pos < fullName.Length && char.IsWhiteSpace(fullName[pos]))
+			{
+				pos++;
+			}
+			int quote = -1;
+			if (pos < fullName.Length && (fullName[pos] == '"' || fullName[pos] == '\''))
+			{
+				quote = fullName[pos++];
+			}
+			for (; pos < fullName.Length; pos++)
+			{
+				char ch = fullName[pos];
+				if (ch == '\\')
+				{
+					if (++pos == fullName.Length)
+					{
+						return false;
+					}
+					ch = fullName[pos];
+					if (ch == '\\')
+					{
+						return false;
+					}
+				}
+				else if (ch == quote)
+				{
+					for (pos++; pos != fullName.Length; pos++)
+					{
+						ch = fullName[pos];
+						if (ch == ',' || ch == '=')
+						{
+							break;
+						}
+						if (!char.IsWhiteSpace(ch))
+						{
+							return false;
+						}
+					}
+					break;
+				}
+				else if (quote == -1 && (ch == '"' || ch == '\''))
+				{
+					return false;
+				}
+				else if (quote == -1 && (ch == ',' || ch == '='))
+				{
+					break;
+				}
+				sb.Append(ch);
+			}
+			value = sb.ToString().Trim();
+			return value.Length != 0 || quote != -1;
+		}
+
+		private static bool TryConsume(string fullName, char ch, ref int pos)
+		{
+			if (pos < fullName.Length && fullName[pos] == ch)
+			{
+				pos++;
+				return true;
+			}
+			return false;
+		}
+
+		private static bool TryParseAssemblyAttribute(string fullName, ref int pos, ref string key, ref string value)
+		{
+			return TryConsume(fullName, ',', ref pos)
+				&& TryParse(fullName, ref pos, out key)
+				&& TryConsume(fullName, '=', ref pos)
+				&& TryParse(fullName, ref pos, out value);
+		}
+
 		internal static ParseAssemblyResult ParseAssemblyName(string fullName, out ParsedAssemblyName parsedName)
 		{
 			parsedName = new ParsedAssemblyName();
 			int pos;
 			ParseAssemblyResult res = ParseAssemblySimpleName(fullName, out pos, out parsedName.Name);
-			if (res != ParseAssemblyResult.OK || pos == fullName.Length)
+			if (res != ParseAssemblyResult.OK)
 			{
 				return res;
 			}
@@ -410,32 +533,30 @@ namespace IKVM.Reflection
 				System.Collections.Generic.Dictionary<string, string> unknownAttributes = null;
 				bool hasProcessorArchitecture = false;
 				bool hasContentType = false;
-				string[] parts = fullName.Substring(pos).Split(',');
-				for (int i = 0; i < parts.Length; i++)
+				while (pos != fullName.Length)
 				{
-					string[] kv = parts[i].Split('=');
-					if (kv.Length != 2)
+					string key = null;
+					string value = null;
+					if (!TryParseAssemblyAttribute(fullName, ref pos, ref key, ref value))
 					{
 						return ParseAssemblyResult.GenericError;
 					}
-					switch (kv[0].Trim().ToLowerInvariant())
+					key = key.ToLowerInvariant();
+					switch (key)
 					{
 						case "version":
 							if (parsedName.Version != null)
 							{
 								return ParseAssemblyResult.DuplicateKey;
 							}
-							if (!ParseVersion(kv[1].Trim(), out parsedName.Version))
-							{
-								return ParseAssemblyResult.GenericError;
-							}
+							parsedName.Version = value;
 							break;
 						case "culture":
 							if (parsedName.Culture != null)
 							{
 								return ParseAssemblyResult.DuplicateKey;
 							}
-							if (!ParseCulture(kv[1].Trim(), out parsedName.Culture))
+							if (!ParseCulture(value, out parsedName.Culture))
 							{
 								return ParseAssemblyResult.GenericError;
 							}
@@ -445,7 +566,7 @@ namespace IKVM.Reflection
 							{
 								return ParseAssemblyResult.DuplicateKey;
 							}
-							if (!ParsePublicKeyToken(kv[1].Trim(), out parsedName.PublicKeyToken))
+							if (!ParsePublicKeyToken(value, out parsedName.PublicKeyToken))
 							{
 								return ParseAssemblyResult.GenericError;
 							}
@@ -455,7 +576,7 @@ namespace IKVM.Reflection
 							{
 								return ParseAssemblyResult.DuplicateKey;
 							}
-							if (!ParsePublicKey(kv[1].Trim(), out parsedName.PublicKeyToken))
+							if (!ParsePublicKey(value, out parsedName.PublicKeyToken))
 							{
 								return ParseAssemblyResult.GenericError;
 							}
@@ -466,7 +587,7 @@ namespace IKVM.Reflection
 							{
 								return ParseAssemblyResult.DuplicateKey;
 							}
-							switch (kv[1].Trim().ToLowerInvariant())
+							switch (value.ToLowerInvariant())
 							{
 								case "yes":
 									parsedName.Retargetable = true;
@@ -484,7 +605,7 @@ namespace IKVM.Reflection
 								return ParseAssemblyResult.DuplicateKey;
 							}
 							hasProcessorArchitecture = true;
-							switch (kv[1].Trim().ToLowerInvariant())
+							switch (value.ToLowerInvariant())
 							{
 								case "none":
 									parsedName.ProcessorArchitecture = ProcessorArchitecture.None;
@@ -514,14 +635,14 @@ namespace IKVM.Reflection
 								return ParseAssemblyResult.DuplicateKey;
 							}
 							hasContentType = true;
-							if (kv[1].Trim().ToLowerInvariant() != "windowsruntime")
+							if (!value.Equals("windowsruntime", StringComparison.OrdinalIgnoreCase))
 							{
 								return ParseAssemblyResult.GenericError;
 							}
 							parsedName.WindowsRuntime = true;
 							break;
 						default:
-							if (kv[1].Trim() == "")
+							if (key.Length == 0)
 							{
 								return ParseAssemblyResult.GenericError;
 							}
@@ -529,20 +650,25 @@ namespace IKVM.Reflection
 							{
 								unknownAttributes = new System.Collections.Generic.Dictionary<string, string>();
 							}
-							if (unknownAttributes.ContainsKey(kv[0].Trim().ToLowerInvariant()))
+							if (unknownAttributes.ContainsKey(key))
 							{
 								return ParseAssemblyResult.DuplicateKey;
 							}
-							unknownAttributes.Add(kv[0].Trim().ToLowerInvariant(), null);
+							unknownAttributes.Add(key, null);
 							break;
 					}
 				}
+				return ParseAssemblyResult.OK;
 			}
-			return ParseAssemblyResult.OK;
 		}
 
-		private static bool ParseVersion(string str, out Version version)
+		internal static bool ParseVersion(string str, out Version version)
 		{
+			if (str == null)
+			{
+				version = null;
+				return true;
+			}
 			string[] parts = str.Split('.');
 			if (parts.Length < 2 || parts.Length > 4)
 			{
@@ -614,9 +740,9 @@ namespace IKVM.Reflection
 			return true;
 		}
 
-		private static bool IsPartial(ParsedAssemblyName name)
+		private static bool IsPartial(ParsedAssemblyName name, Version version)
 		{
-			return name.Version == null || name.Culture == null || name.PublicKeyToken == null;
+			return version == null || name.Culture == null || name.PublicKeyToken == null;
 		}
 
 		private static bool IsStrongNamed(ParsedAssemblyName name)
