@@ -20,8 +20,6 @@
 // Copyright (c) 2007 Novell, Inc.
 //
 
-#if NET_2_0
-
 using System.ComponentModel;
 using System.Collections;
 using System.Collections.Generic;
@@ -194,6 +192,36 @@ namespace System.Windows.Forms {
 			ResetBindings (true);
 		}
 
+		private void ConnectDataSourceEvents (object dataSource)
+		{
+			if (dataSource == null)
+				return;
+			
+			ICurrencyManagerProvider currencyManagerProvider = dataSource as ICurrencyManagerProvider;
+			if (currencyManagerProvider != null && currencyManagerProvider.CurrencyManager != null) {
+				currencyManagerProvider.CurrencyManager.CurrentItemChanged += OnParentCurrencyManagerChanged;
+				currencyManagerProvider.CurrencyManager.MetaDataChanged += OnParentCurrencyManagerChanged;
+			}
+		}
+
+		private void OnParentCurrencyManagerChanged (object sender, EventArgs args)
+		{
+			// Essentially handles chained data sources (e.g. chained BindingSource)
+			ResetList ();
+		}
+
+		private void DisconnectDataSourceEvents (object dataSource)
+		{
+			if (dataSource == null)
+				return;
+
+			ICurrencyManagerProvider currencyManagerProvider = dataSource as ICurrencyManagerProvider;
+			if (currencyManagerProvider != null && currencyManagerProvider.CurrencyManager != null) {
+				currencyManagerProvider.CurrencyManager.CurrentItemChanged -= OnParentCurrencyManagerChanged;
+				currencyManagerProvider.CurrencyManager.MetaDataChanged -= OnParentCurrencyManagerChanged;
+			}
+		}
+
 		void IBindingListChangedHandler (object o, ListChangedEventArgs args)
 		{
 			if (raise_list_changed_events)
@@ -316,10 +344,12 @@ namespace System.Windows.Forms {
 			get { return datasource; }
 			set {
 				if (datasource != value) {
-					if (datasource == null)
+					if (value == null)
 						datamember = String.Empty;
 
+					DisconnectDataSourceEvents (datasource);
 					datasource = value;
+					ConnectDataSourceEvents (datasource);
 					ResetList ();
 
 					OnDataSourceChanged (EventArgs.Empty);
@@ -925,7 +955,7 @@ namespace System.Windows.Forms {
 		public void ResetBindings (bool metadataChanged)
 		{
 			if (metadataChanged)
-				OnListChanged (new ListChangedEventArgs (ListChangedType.PropertyDescriptorChanged, 0));
+				OnListChanged (new ListChangedEventArgs (ListChangedType.PropertyDescriptorChanged, null));
 
 			OnListChanged (new ListChangedEventArgs (ListChangedType.Reset, -1, -1));
 		}
@@ -1039,5 +1069,3 @@ namespace System.Windows.Forms {
 		}
 	}
 }
-
-#endif
